#use strict;
use warnings;
use diagnostics;

use ChIPSeqPipeline_sample;
use ChIPSeqPipeline_group;
use ChIPSeqPipeline_comparison;

package Study;

sub new {
  my $self  = {};
  $self->{ACCESSION} = undef;
  $self->{PARAM_FILE} = undef;
  $self->{GENOME_VERSION}  = undef;
  $self->{SAMPLES_DIR_PATH}  = undef;
  $self->{STUDY_MNGR}  = undef;
  $self->{PROCESS_MNGR}  = undef;
  $self ->{SAMPLES} = [];
  $self ->{GROUPS} = [];
  $self ->{COMPARISONS} = [];
  $self ->{UNPROCESSED_SAMPLES} = [];
  $self ->{UNPROCESSED_GROUPS} = [];
  $self ->{UNPROCESSED_COMPARISONS} = [];
  bless($self); 
  return $self;
}

sub accession {
  my $self = shift;
  if (@_) { $self->{ACCESSION}  = shift; }
  return $self->{ACCESSION};
}
sub paramFile {
  my $self = shift;
  if (@_) { 
    $self->{PARAM_FILE}  = shift;
    $self->parseParams();
  }
  return $self->{PARAM_FILE};
}
sub samplesDirPath {
  my $self = shift;
  if (@_) { $self->{SAMPLES_DIR_PATH}  = shift; }
  return $self->{SAMPLES_DIR_PATH};
}
sub study_mngr {
  my $self = shift;
  if (@_) { $self->{STUDY_MNGR}  = shift; }
  return $self->{STUDY_MNGR};
}
sub genomeVersion {
  my $self = shift;
  if (@_) { $self->{GENOME_VERSION}  = shift; }
  return $self->{GENOME_VERSION};
}
sub groups{
  my $self = shift;
  if(@_){ push@{ $self->{GROUPS} }, @_ }
  return @{ $self->{GROUPS} };
}
sub samples{
  my $self = shift;
  if(@_){ push@{ $self->{SAMPLES} }, @_ }
  return @{ $self->{SAMPLES} };
}
sub comparisons{
  my $self = shift;
  if(@_){ push@{ $self->{COMPARISONS} }, @_ }
  return @{ $self->{COMPARISONS} };
}
sub groupsNum{
  my $self = shift;
  return ( $#{ $self->{GROUPS} } +1 );
}
sub comparisonsNum{
  my $self = shift;
  return ( $#{ $self->{COMPARISONS} } + 1 );
}

sub parseParams{
  ### This sub takes a study param file and create samples and groups
  ### Input is a parameter file 3 parts seperated by a line starting with "----".
  ### Part 1: Key val (genome version)
  ### Part 2:  Id Title	Platform Description	Group (tab delimited)
  ### Part 3:  Comparison   Trat	Ctrl	Title (tab delimited)
  ### Return valuse is number of comparisons (later to become biosets)
  my $self = shift;
  if( !defined($self->{PARAM_FILE}) ) { print "ERROR: parameter file not initiated\n"; return 0; }
  my $paramFile =  $self->{PARAM_FILE};
  print"------------- param file -------------------\n";
  my $section = 1; my $firstLine = 1;
  open PARAMFILE, "<","$paramFile" or print"ERROR: Could not open param file $paramFile" ;
  while (my $line = <PARAMFILE>){
    #print"--- section $section first $firstLine line : $line\n";
    if( $line =~ /^#/ ){ next; }### Skip comment lines
    if( length($line) < 2 ){ next; }### Skip empty lines
    if($firstLine == 1){ $firstLine = 0; next; }
    if( $line =~ m/^-----/ ){ $section++; $firstLine = 1; next; }
    chomp $line;
    my @lineArr =  split( /\t/, $line );
    #my $paramNum = $#lineArr;
    if( $section == 1 ){ $self -> parseSection1_keyVal( @lineArr ) }
    if( $section == 2 ){ $self -> parseSection2_samplesGroups( @lineArr ) }
    if( $section == 3 ){ $self -> parseSection3_comparisons( @lineArr ) }
  }
  print"------------- End param file ----------------\n";
}


sub parseSection1_keyVal{
  my $self = shift;
  my @lineArr = @_;
  my $paramNum = $#lineArr;
  if($paramNum != 1){ print"ERROR: Format error in study parameter file [$self->paramFile()].\nParam number is $paramNum. Expecting 2 parameters (key and val sererated by a tab) in section 1.\nLine: @lineArr\n"; return 0;}
  if( $lineArr[0] eq "GENOME_VERSION" ){ $self-> genomeVersion( $lineArr[1] ); }
  if( $lineArr[0] eq "ACCESSION" ){ $self-> accession( $lineArr[1] ); }
  if( $lineArr[0] eq "SAMPLES_DIR_PATH" ){ $self-> samplesDirPath( $lineArr[1] ); }
}

sub parseSection2_samplesGroups{
  my $self = shift;
  my @lineArr = @_;
  my $paramNum = $#lineArr;
  if($paramNum != 4){ print"ERROR: Format error in study parameter file [$self->paramFile()].\nParam number is $paramNum. Expecting 5 parameters in section 2.\nExpecting the following header:\nId\tTitle\tPlatform\tDescription\tGroup\nEach sample should be a seperte line\n"; return 0;}
  my ($id, $sampleTitle, $platform, $desc, $group) = @lineArr;
  #print"*** section 2: $id, $sampleTitle, $platform, $desc, $group\n";
  my $sampleObjct = $self -> buildSample( $id, $sampleTitle, $platform, $desc );
  $self -> samples( $sampleObjct ); 
  ### Deal with multiple groups per sample:
  my @groupArr = (); 
  if( $group =~ m/,/ ){ @groupArr =  split( /,/, $group ); }
  else{ push( @groupArr, $group ); }### Assign a single group name
  foreach my $currGroup (@groupArr){
    my $groupObjct = $self -> buildGroup( $group );
    $self -> linkSamplesGroup( $sampleObjct, $groupObjct );
  }
}

sub parseSection3_comparisons{
  my $self = shift;
  my @lineArr = @_;
  my $paramNum = $#lineArr;
  if($paramNum != 3){ print"ERROR: Format error in study parameter file [$self->paramFile()].\nParam number is $paramNum. Expecting 4 parameters in section 3.Expecting the following header:\nComparison\tTrat\tCtrl\tTitle\nEach comparison should be a seperte line\n"; return 0;}
  my ($cmpNum, $treatment, $control, $cmpTitle) = @lineArr;
  #print"*** section 3: $cmpNum, $treatment, $control, $cmpTitle\n";
  my $comparisonObjct = $self -> buildComparison( $cmpNum, $treatment, $control, $cmpTitle );
  print"Add to cmp $cmpNum, $treatment, $control, $cmpTitle\n";
  $self -> comparisons( $comparisonObjct );
} 

sub buildSample{
  ### Creates a sample object by name, title, platform and description. Groups or files are not assigned at this point.
  my $self = shift;
  my( $id, $sampleTitle, $platform, $desc ) = @_;
  my $sampleObjct = Sample->new;
  $sampleObjct -> name( $id ); $sampleObjct -> title( $sampleTitle ); $sampleObjct -> platform( $platform ); $sampleObjct -> desc( $desc );
  return $sampleObjct;
}

sub buildGroup{
  ### Creates a new group object and assigns name. If group is new (does not already appear in groups_names array) it is added to group array. Returns old or new group object.
  my $self = shift;
  if(!(@_) ){ print"Error: GRoup name must be supplied to build a group\n"; }
  my( $groupName ) = @_;
  my $groupObjct = Group->new;
  $groupObjct -> study( $self );
  $groupObjct -> name ($groupName);
  my $isNewGroupRes = $self -> isNewGroup( $groupName );
  if( $isNewGroupRes == 1 ){
    $self -> groups( $groupObjct ); 
  }
  else{ $groupObjct = $isNewGroupRes; }
  return $groupObjct;
}

sub isNewGroup{
  ### Returns 1 if group is new. Returns the group object is group is found in groups array. If no groups exist the input group is considered new.
  my $self = shift;
  my($inputGroupName) = @_;
  #print"checking group name [$inputGroupName]\n";###Repeats too many times
  my $gn = $self -> groupsNum(); 
  #print" group num = $gn $inputGroupName [$inputGroupName] \n";
  if ( $self -> groupsNum() == 0 ){ return 1; }
  foreach my $group ( $self -> groups() ) {
    #my $gname = $group->name(); print"compare to group name [$gname]\n";
    if( $inputGroupName eq $group->name() ){ return $group; }
  }
  #print"new group\n";
  return 1;
}

sub linkSamplesGroup{
  my $self = shift;
  my($sampleObjct, $groupObjct) = @_;
  $sampleObjct -> addGroup( $groupObjct );
  $groupObjct -> addSample( $sampleObjct );
}

sub buildComparison{
  my $self = shift;
  my($cmpNum, $treatment, $control, $cmpTitle) = @_;
  my $comparisonObjct;
  if( $self -> isNewComparison($treatment, $control) == 1 ){
    $comparisonObjct = Comparison->new;
    $comparisonObjct -> name( $cmpNum ); $comparisonObjct -> title( $cmpTitle );  
    ###Find group objetc by its name for treatment and control
    my $treatmentObjct = $self -> isNewGroup( $treatment );
    my $controlObjct = $self -> isNewGroup( $control );
    my $isGRoupsExist = $self -> checkGroupsExist( $treatmentObjct, $controlObjct, $cmpNum, $treatment, $control, $cmpTitle );
    if( $isGRoupsExist == 0 ){ $comparisonObjct -> status( "MISSING_GROUP" ); }
    $comparisonObjct -> treatment( $treatmentObjct ); $comparisonObjct -> control( $controlObjct );
  }
  return $comparisonObjct;
}

sub checkGroupsExist{
  ### Verify treatmetn and control objects exist in group array (by isNewGroup parameter - currently group name). If not - provide an error message. Returns 1 if groups both exist and 0 if any of them cannot be found.
  my $self = shift;
  my($treatmentObjct,$controlObjct, $cmpNum, $treatment, $control, $cmpTitle) = @_;
  if( $treatmentObjct == 1 || $controlObjct == 1 ){ 
    print"Error: Comparison $cmpNum titled [$cmpTitle] contains :\n"; 
    if( $treatmentObjct == 1 ){ print "\ta treatment name [$treatment] without a coresponding group\n"; }
    if( $controlObjct == 1 ){ print "\ta control name [$control] without a coresponding group\n"; }
    my $paramFile = $self -> paramFile(); my $study = $self -> accession();
    print"Please check your parameter file  $paramFile for study $study\n"; 
    return 0; 
  }
  return 1;
}

sub isNewComparison{
  ### Returns 1 if group is new. Returns the comparison object is group is found in comparisons array. If no comparisons exist the input comparison is considered new.
  my $self = shift;
  my($treatment, $control) = @_;
  if ( $self -> comparisonsNum() == 0 ){ return 1; }
  foreach my $cmp ( $self -> comparisons() ){
    if( $cmp->treatment()->name() eq $treatment && $cmp->control()->name() eq $control ){ return $cmp; }
  }
  return 1;
}

sub print{
   my $self = shift;
   my $cmpNum = $self->comparisonsNum(); my $studyAccession =  $self->accession();
   if( !(defined($cmpNum)) || !(defined($studyAccession)) ){ print"Error: Undefined parameter for study\nPlease check your input file list or the content of your parameter file directory\n"; return; }
  print"\t~~~~~~~~~~ Study  $studyAccession has $cmpNum comparisons ~~~~~~~~~~~\n";
  foreach my $cmp ( $self -> comparisons() ){
    $cmp->print();
  }
}
1;
